""" generate a xmi-uml database and an argouml project file
if a diagrams definiton is given (see diadef.dtd), the matching pgml file are
generated
"""
__revision__ = '$Id: pyargo.py,v 1.6 2002/01/28 09:57:36 syt Exp $'

import os
from xmlconf import DictSaxHandlerMixin, PrefReader
from base import IdGeneratorMI
    
## saxHandler definition #######################################################
class DiaDefsSaxHandler(DictSaxHandlerMixin):
    """ definition of the structure of the diadef file, which will enable the MI 
    to fill the DiaDefs dictionary
    """
    _MASTER = {}
    _S_LIST = ('class-diagram', 'package-diagram', 'state-diagram',
               'class', 'package')
    _LIST = ()
    _KEY = ('owner', 'name', 'include')
    _CB = {}

_BASE = 'uci.uml.visual'
#_BASE = 'org.argouml.uml.diagram.static_structure.ui'

## utility functions ###########################################################
class ArgoWriter(IdGeneratorMI):
    def __init__(self, project, diadef, idstart, uuid_f):
        self.initCounter(idstart)
        self._p = project
        self._d = diadef
        self.f_uuid = uuid_f
        self._files = []
        
    def write(self, xmi):
        self.write_pgml()
        self.write_argo(xmi)
        
    def write_pgml(self):
        diadefs = self._d
        if diadefs:
            for diadef in diadefs.get('class-diagram', []):
                self._write_class_diagram(diadef)
            for diadef in diadefs.get('package-diagram', []):
                self._write_package_diagram(diadef)
            
    def write_argo(self, xmi):
        """ write to a file an argo project description """
        p_f_name = self._p.name.replace(' ', '_')
        info('writing XMI file to %s.xmi' % p_f_name)        
        xmi_file = '%s.xmi' % p_f_name
        xmi_f = open(xmi_file, 'w+')
        from xml.dom.ext import PrettyPrint
        PrettyPrint(xmi, xmi_f)
        info('writing Argo project file to %s.argo' % p_f_name)
        argo_file = '%s.argo' % p_f_name
        argo_f = open(argo_file, 'w+')
        from logilab.pyreverse import version, modname
        print >>argo_f,'''<?xml version="1.0" encoding="ISO-8859-1" ?>
<!DOCTYPE argo SYSTEM "argo.dtd" >
<argo>
  <documentation>
    <authorname>%s</authorname>
    <version>%s</version>
    <description>
Generated by pyargo tools
    </description>
  </documentation>
  <searchpath href="PROJECT_DIR" />
  <member type="xmi" name="%s.xmi"
  />'''% (modname, version, self._p.name)
        for file in self._files:
            print >>argo_f, '''
  <member type="pgml" name="%s"/>'''%file
        print >>argo_f,'''
  <historyfile name=""/>
  <stats/>
</argo>'''
        self._files.append(xmi_file)
        self._files.append(argo_file)

    def zargo(self, zipper='zip'):
        assert self._files
        p_f_name = self._p.name.replace(' ', '_')
        info('zipping all files to %s.zargo' % p_f_name)
        cmd = '%s %s.zargo' % (zipper, p_f_name)
        for f in self._files:
            cmd = '%s %s' % (cmd, f)
        os.system(cmd)
##         import zipfile
##         zip = zipfile.ZipFile('%s.zargo'%p_name, 'w',
##                               zipfile.ZIP_DEFLATED)
##         zip.write(a_file)
##         zip.write(x_file)
##         for f in files:
##             zip.write(f)
##         zip.close()
        
    def clean(self):
        for f in self._files:
            os.remove(f)
        self._files = []

    def _write_package_diagram(self, diadef):
        f = self._open_pgml('UMLClassDiagram',
                            diadef.get('name', 'No name %d'%self.id_count))
        packages = []
        classes = self._resolve_packages(diadef, packages)
        self._write_packages(f, packages)
        self._write_classes(f, classes, packages)
        print >>f, '</pgml>'

    def _write_class_diagram(self, diadef):
        """ write to a file a class diagram according to the pgml dtd """
        f = self._open_pgml('UMLClassDiagram',
                            diadef.get('name', 'No name %d'%self.id_count))
        objects = []
        self._resolve_classes(diadef, objects)
        self._write_classes(f, objects)
        print >>f, '</pgml>'

    def _open_pgml(self, type, name):
        file = '%s.pgml' % name.replace(' ', '_')
        info('creating diagram %s'%file)
        f = open(file, 'w+')
        print >>f,'''<?xml version = "1.0" encoding = "ISO-8859-1" ?>
<!DOCTYPE pgml SYSTEM "pgml.dtd" >
<pgml description="uci.uml.visual.%s|%s" name="%s">'''%(
            type, self.f_uuid(self._p.id), name)
        self._files.append(file)
        return f

    def _resolve_packages(self, diadef, list):
        classes = [] 
        self._resolve_classes(diadef, classes)
        for package in diadef['package']:
            name = package['name']
            c = self._p.get_module(name)
            if c is None:
                print >>sys.stderr, 'Warning: no module named %s'%name
                print >>sys.stderr, c
                continue
            if package.has_key('include'):
                if package['include'] == 'all':
                    for classe in c.childs:
                        if classe.TYPE == 'class':
                            classes.append(classe)
                elif package['include'] != 'no':
                    self._resolve_classes(package, classes)
            else:
                self._resolve_classes(package, classes)
            list.append(c)
        return classes

    def _resolve_classes(self, diadef, list):
        for klass in diadef.get('class', []):
            name, module = klass['name'], klass.get('module', '')
            c = self._p.get_classes(name, module)
            if len(c) > 1:
                print >>sys.stderr, 'Warning: more than one class named %s'%name
                print >>sys.stderr, c
                c = c[0]
            elif len(c) == 0:
                print >>sys.stderr, 'Warning: no class named %s'%name
                print >>sys.stderr, c
                continue
            else: c = c[0]
            list.append(c)

    def _write_packages(self, f, objects):
        for o in objects:
            o._fig_id = self.generateId()
            print >>f, group("FigPackage", o._fig_id, self.f_uuid(o.id))
        ## package dependencies
        for o in objects:
            o1 = o._fig_id
            for p in o.get_deps():
                if p in objects:
                    o2 = p._fig_id
                    type = 'FigDependency'
                    id = self.generateId()
##                        print >>f, group(type, id, self.f_uuid(id), '''
                    print >>f, group(type, id, None, '''
      sourcePortFig="Fig%s.0"
      destPortFig="Fig%s.0"
      sourceFigNode="Fig%s"
      destFigNode="Fig%s"'''%(o1, o2, o1, o2), fill=0)

    def _write_classes(self, f, objects, parents=[]):
        for o in objects:
            m = o.get_module()
            if m in parents:
                private = 'enclosingFig="Fig%s"'%m._fig_id
            else: private = ''
            if o.type == 'interface':
                type = 'FigInterface'
            else:
                type = 'FigClass'
            o._fig_id = self.generateId()
            print >>f, group(type, o._fig_id, self.f_uuid(o.id), private)
        links = {}
        ## inheritance link
        for o in objects:
            o1 = o._fig_id
            for p in o.parents:
                if p in objects:
                    o2 = p._fig_id
                    if p.type == 'interface':
                        type = 'FigRealization'
                        fill = 0
                    else:
                        type = 'FigGeneralization'
                        fill = 1
                    id = self.generateId()
##                    print >>f, group(type, id, self.f_uuid(id), '''
                    print >>f, group(type, id, None, '''
      sourcePortFig="Fig%s.0"
      destPortFig="Fig%s.0"
      sourceFigNode="Fig%s"
      destFigNode="Fig%s"'''%(o1, o2, o1, o2), fill)
            ## prepare associations
            for attr in o.childs:
                if attr.TYPE == 'attribute':
                    try:
                        if attr.default in objects:
                            links.setdefault(o, {})[attr.default] = 1
                            links.setdefault(attr.default, {})[o] = 1
                    except:
                        continue
        ## generate associations
        for link, ends in links.items():
            o1 = link._fig_id
            for end in ends.keys():
                o2 = end._fig_id
                id = self.generateId()
##                print >>f, group("FigAssociation", id, self.f_uuid(id),'''
                print >>f, group("FigAssociation", id, None, '''
      sourcePortFig="Fig%s.0"
      destPortFig="Fig%s.0"
      sourceFigNode="Fig%s"
      destFigNode="Fig%s"'''%(o1, o2, o1, o2))
                del links[end][link]
                del links[link][end] 


def read_diadefs_file(file):
    """ read a diadef file and return the DiaDef dictionnary"""
    info('reading diagrams definitions file')
    diadefs = {} 
    reader = PrefReader(DiaDefsSaxHandler, diadefs)
    try:
        reader.fromFile(file)
    except:
        import traceback
        traceback.print_exc()
        print >>sys.stderr, 'error while reading definitions file %s' % file
    return diadefs

def make_diadefs(project):
    """
    _S_LIST = ('class-diagram', 'package-diagram', 'state-diagram',
               'class', 'package')
    _LIST = ()
    _KEY = ('owner', 'name', 'include')
    """
    info('making default diagrams')
    diadefs = {}
    if len(project.childs) > 1:
        diadefs['package-diagram'] = [{'name' : 'packages %s'%project.name,
                                       'package': []}]
        diadefs['class-diagram'] = [{'name' : 'classes %s'%project.name,
                                     'class': []}]
        for m in project.childs:
            diadefs['package-diagram'][0]['package'].append({'name' : m.name})
            for c in m.childs:
                if c.TYPE == 'class':
                    diadefs['class-diagram'][0]['class'].append({'name' : c.name,
                                                                 'owner': m.name})
    else:
        diadefs['class-diagram'] = [{'name' : 'classes %s'%project.name,
                                     'class': []}]
        for m in project.childs:
            for c in m.childs:
                if c.TYPE == 'class':
                    pass
    return diadefs

def group(descr, id, idref, private='', fill=1):
    if fill: fill = '''
       fill="1"
       fillcolor="-1"'''
    else: fill = ''
    if idref: idref = '''
       href="%s"''' % idref
    else: idref = ''
    return '''
  <group name="Fig%d"
       description="%s.%s"%s%s
       stroke="1"
       strokecolor="-16777216">
    <private>%s
    </private>
  </group>'''%(id, _BASE, descr, idref, fill, private)


def info(msg):
    print '-'*80
    print msg

    
## main function ###############################################################
USAGE = """
pyargo parse python sources files and extract an ArgoUML project file from them

Usage: pyargo [options] <python-file>...
       pyargo [options] -x/--xmi <xmi-file>
Options: _ -h/--help, display this help message and exit
         _ -p/--project <name>, set project name to <name>
         _ -d/--diadefs <diadefs-file>, create diagram according to the diagrams
                                        definitions in <file>
         _ -f/--filter-mode <mode>, filter attributes and functions according to
                                    <mode>. correct modes are
                                    _ 'ALL' no filter (default)
                                    _ 'PUB_ONLY' filter all non public attributes
                                    (equivalent to PRIVATE+SPECIAL_A)
                                    _ 'SPECIAL' filter Python special functions
                                    except constructor
                                    _ 'SPECIAL_A' filter Python special functions
                                    and constructor
                                    _ 'PRIVATE' filter protected and private
                                    attribute
         _ -z/--zargo, compress output (ArgoUML > 0.9)
                       require to have 'zip' in your path
         _ -w/--zipper <pgm name>, compress output using <pgm name> to zip all files
                                   require to have <pgm name> in your path
                                   default zipper i 'zip'
"""

import sys

def Run(args):
    import getopt
    import xmi_uml
    l_opt = ['help', 'project=', 'diadefs=', 'filter-mode=', 'zargo'
             'zipper=']#, 'xmi=']
    (opt, args) = getopt.getopt(args, 'hp:d:f:zw:', l_opt)
    p_name, diadefs_file, dest, filter, compress = 'No Name', None, sys.stdout, 0, 0
    zipper = 'zip'
    for name, val in opt:
        if name == '-h' or name == '--help':
            print >>sys.stderr, USAGE
            return
        elif name == '-p' or name == '--project':
            p_name = val
        elif name == '-d' or name == '--diadefs':
            diadefs_file = val
        elif name == '-f' or name == '--filter-mode':
                mode = 0
                for mn in  val.split('+'):
                    try:
                        mode += eval('xmi_uml.%s'%mn)
                    except:
                        import traceback
                        traceback.print_exc()
                        print >>sys.stderr, 'Unknown filter mode !!!'
                        print >>sys.stderr, USAGE
                        return
        elif name == '-z' or name == '--zargo':
            compress = 1
        elif name == '-w' or name == '--zipper':
            zipper = val
##        elif name == '-x' or name == '--xmi':
##            xmi = val
    if not args:
        print >>sys.stderr, 'No file to parse !!!'
        print >>sys.stderr, USAGE
        return
    ## extract project representation
    info('parsing project')
    from pyparser import PyParser
    parser = PyParser()
    p = parser.parse_all(args, p_name, lambda f,m: sys.stdout.write(
        '  _ parsing file %s\n'%f))
    p.resolve_hierarchy()
    p.resolve_imports()
    ## transform into xmi-uml
    info('generating XMI UML')
    k = xmi_uml.Klass2DomXmi()
    d = k.parse(p, parser.id_count, filter)
    ##  read and interpret diagram definitions
    if diadefs_file is not None:
        diadefs = read_diadefs_file(diadefs_file)
##        from pprint import pprint
##        pprint(diadefs) 
    else:
        diadefs = make_diadefs(p)  
    w = ArgoWriter(p, diadefs, k.id_count, k.formatUUID)
    w.write(d.ownerDocument)
    ## write argo project file
    if compress:
        w.zargo(zipper)
        w.clean()

if __name__ == "__main__":
    Run(sys.argv[1:])

    

