""" print useful statistic about the project """

__revision__ = '$Id: pystats.py,v 1.4 2002/01/14 11:15:47 syt Exp $'

from base import AbstractBase

def get_stats(node, sysmod={}, data=None, verbose=0):
    """ recursivly calculate statistic from project description """
    if data is None:
        data = {'stat' : {},
                'unresolved' : {},
                'import*' : {},
                'cycle' : []}
        if verbose:
            print '-'*80
            print '                        PARSE PROJECT'
            _separ()
    if node.TYPE in ('function', 'class', 'module'):
        data['stat'].setdefault(node.TYPE, [0, 0])
        data['stat'][node.TYPE][0] += 1
        if hasattr(node, 'doc') and node.doc:
            data['stat'][node.TYPE][1] += 1
        elif verbose:
            if node.name == '__init__': s = '%s.%s'%(node.fullname(), node.name)
            else: s = node.fullname()
            print '%s %s is not documented'%(node.TYPE, s)
        for mod, asmod in node.imports:
            #print asmod, type(mod)
            if asmod == '*':
                if verbose:
                    print 'import * in %s' % node.fullname
                name=node.fullname()
                data['import*'].setdefault(name, 0)
                data['import*'][name] += 1
            if not isinstance(mod, AbstractBase):
                if not sysmod.has_key(mod.split('.')[0]):
                    data['unresolved'].setdefault(mod, []).append(node.fullname())
        if node.TYPE == 'module':
            for dep in node.get_deps():
                if node in dep.get_deps():
                    data['cycle'].append((node, dep))
                
    for c in node.childs:
        get_stats(c, sysmod, data, verbose)
    return data
    
def print_stats(project, data):
    """ nicely format statistics results """
    print '-'*80
    print '                        STATISTICS FOR PROJECT %s'%project.name
    _separ()
    print 'A few numbers:\n'
    percent = (float(data['stat']['module'][1])/data['stat']['module'][0])*100
    data['stat']['module'] += [percent]
    print ' %d modules (<%d / %d%% documented)'%tuple(data['stat']['module'])
    percent = (float(data['stat']['class'][1])/data['stat']['class'][0])*100
    data['stat']['class'] += [percent]
    print ' %d classes (<%d / %d%% documented)'%tuple(data['stat']['class'])
    percent = (float(data['stat']['function'][1])/data['stat']['function'][0])*100
    data['stat']['function'] += [percent]
    print ' %d functions (<%d / %d%% documented)\n'%tuple(data['stat']['function'])
    _separ()    
    new = {}
    for mod, files in data['unresolved'].items():
        prefixes = mod.split('.')
        d = (new, ())
        for prefix in prefixes:
            d = d[0].setdefault(prefix, [{}, []])
        d[1] += files
    if new:
        print 'Unresolved external imports:\n'
        print repr_defs(new)
    else:
        print 'No unresolved external imports 8-)\n'
    _separ()
    if data['import*']:
        print 'Bad guys which imports *:\n'
        for file, num in data['import*'].items():
            print ' %s (%s)'% (file, num)
    else:
        print 'Good, no files imports * !!\n'
    _separ()
    if data['cycle']:
        print 'Cyclic dependencies:\n'
        done = []
        for _from, to in data['cycle']:
            if not (_from, to) in done:
                done.append((to, _from))
                print ' from %s to %s'% (_from.name, to.name)
    else:
        print 'No cyclic dependencies\n'
    print '-'*80

def repr_defs(data, indent_str=None):
    """ return a string which represents the defs dictionary """
    s = ''
    l = data.items()
    for i in range(len(l)):
        mod, (sub, files) = l[i]
        if not files: files = ''
        if indent_str is None:
            s = '%s\n%s %s\n' % (s, mod, files)
            sub_indent_str = '  '
        else:
            s = '%s%s\-%s %s\n' % (s, indent_str, mod, files)
            if i == len(l)-1:
                sub_indent_str = '%s  ' % indent_str
            else: sub_indent_str = '%s| ' % indent_str
        s = '%s%s'% (s, repr_defs(sub, sub_indent_str))
    return s

def _separ():
    """ print a separator """
    print '%s%s\n'%(' '*10, '-'*60)

USAGE = """
pystats parse python sources files and extract some statistics and a tree of \
imported modules

Usage: pystats [options] <file>...

Options: _ -h/--help, display this help message and exit
         _ -p/--project <name>, set project name to <name>
         _ -v/--verbose, print the maximum information
"""
import sys

def Run(args):
    import getopt
    (opt, args) = getopt.getopt(args, 'hvp:', ['help', 'verbose', 'project='])
    project, verbose = '', 0
    for name, val in opt:
        if name == '-h' or name == '--help':
            print >>sys.stderr, USAGE
            return
        elif name == '-p' or name == '--project':
            project = val
        elif name == '-v' or name == '--verbose':
            verbose = 1
    from pyparser import PyParser
    parser = PyParser()
    p = parser.parse_all(args, project)
    p.resolve_imports()
#    from pprint import pprint
#    pprint(p._defs)
    ## load file with standards modules
    from data import standard_21
    mods = {}
    for mod in standard_21.MODS:
        mods[mod] = 1
    stats = get_stats(p, mods, verbose=verbose)
    print_stats(p, stats)
    
if __name__ == "__main__":
    Run(sys.argv[1:])

    
