#!/usr/bin/env python
""" convert the class hierarchy describing the project to XMI 1.0 UML 1.3
DOM representation
please notices that it is based on file generated by ArgoUML
"""
__revision__ = "$Id: xmi_uml.py,v 1.5 2002/01/15 12:01:47 syt Exp $"

try:
    from xml.dom import EMPTY_NAMESPACE as NO_NS
except:
    NO_NS = None
from base import Klass, AbstractBase, IdGeneratorMI

_CORE = 'Foundation.Core.'
MAX_CHAR = 20

## filter modes
_CONSTRUCTOR = 1
_SPECIAL = 2
_PROTECTED = 4
_PRIVATE = 8

ALL = 0
PUB_ONLY = _CONSTRUCTOR + _SPECIAL + _PROTECTED + _PRIVATE
SPECIAL = _SPECIAL
SPECIAL_A = _SPECIAL + _CONSTRUCTOR
PRIVATE = _PROTECTED + _PRIVATE

class Klass2DomXmi(IdGeneratorMI):
    """ parser class """
    
    def parse(self, project, idstart, filter_mode=ALL, doc=None):
        """ initialize and parse a project hierarchy, with id counter started
        at idstart ; optionaly use doc as DOM DocumentNode
        """
        self._uuid = '127-0-0-1-unbelid:entifiant:'
        self.initCounter(idstart)
        self._xref_tab = {}
        self._filter = filter_mode
        if doc is None:
            from xml.dom.minidom import DOMImplementation 
            di = DOMImplementation()
            doctype = di.createDocumentType('XMI', None, 'uml13.dtd')
            self._doc = di.createDocument(NO_NS, 'XMI', doctype=doctype)
        else: self._doc = doc
        root, content = self.open_project(project)
        self._stack = [content]
        self._parse(project)
        return root

    def _parse(self, node):
        """ parse the project tree """
        signals, modules = {}, {}
        attributes = []
        doc = self._doc
        ## process each module
        for module in node.childs:
            m, content = self.open_package(module)
            self._stack[-1].appendChild(m)
            self._stack.append(content)
            signals[content] = []
            modules[module] = m
            ## process classes
            for child in module.childs:
                if child.TYPE == 'class':
                    if not child.signal == 'true':
                        n = self.open_class_def(child)
                        content.appendChild(n)
                    else: signals[content].append(child)
                    for n in child.childs:
                        if n.TYPE=='attribute' and hasattr(n, 'default') and\
                           isinstance(n.default,Klass):
                            attributes.append(n)
            del self._stack[-1]
        ## resolve modules dependences
        delayed = []
        for module, dom_n in modules.items():
            deps = module.get_deps()
            for dep in deps:
                d, id = self.open_dependency(module, dep)
                before=dom_n.lastChild
                while before.previousSibling.nodeName == '%sModelElement.supplierDependency'%_CORE:
                    before = before.previousSibling
                d = add_me_dependency(dom_n, client=1, before=before)
                add_idref(d, '%sDependency'%_CORE, id)
                delayed.append
                d = add_me_dependency(modules[dep], before=modules[dep].lastChild,
                                      client=0)
                add_idref(d, '%sDependency'%_CORE, id)
        ## process generalization
        for c, p in self._xref_tab.values():
            p.appendChild(c)
        ## process exceptions
        for parent, childs in signals.items():
            for child in childs:
                n = self.open_class_def(child)
                if hasattr(child, 'context'):
                    for c in child.context:
                        me = doc.createElementNS(
                            NO_NS, 'Behavioral_Elements.Common_Behavior.Signal.context')
                        n.appendChild(me)
                        add_idref(me, '%sBehavioralFeature'%_CORE, c.id)
                        parent.appendChild(n)
        ## process associations
        for child in attributes:
            n = self.open_association(child)
            self._stack[-1].appendChild(n)
            
    def open_dependency(self, module, dep):
        id = self.generateId()
        d = id_element(self._doc, '%sDependency'%_CORE, id, self._stack[0])
        add_value(d, '%sModelElement.isSpecification'%_CORE, 'false')
        c = add_dependency(d, client=1)
        add_idref(c,'%sModelElement'%_CORE, module.id)
        c = add_dependency(d, client=0)
        add_idref(c,'%sModelElement'%_CORE, dep.id)
        return d, id
    
    def open_project(self, project):
        """ initialize an xmi dom tree """
        doc = self._doc
        root = self._doc.documentElement #doc.createElementNS(NO_NS, 'XMI')
        doc.appendChild(root)
        root.setAttributeNS(NO_NS, 'xmi.version', '1.0')
        n1 = doc.createElementNS(NO_NS, 'XMI.header')
        root.appendChild(n1)
        docu = doc.createElementNS(NO_NS, 'XMI.documentation')
        n1.appendChild(docu)
        from logilab.pyreverse import version, modname
        n = doc.createElementNS(NO_NS, 'XMI.exporter')
        n.appendChild(doc.createTextNode(modname))
        docu.appendChild(n)
        n = doc.createElementNS(NO_NS, 'XMI.exporterVersion')
        n.appendChild(doc.createTextNode(version))
        docu.appendChild(n)
        meta = doc.createElementNS(NO_NS, 'XMI.metamodel')
        n1.appendChild(meta)
        meta.setAttributeNS(NO_NS, 'xmi.name', 'UML')
        meta.setAttributeNS(NO_NS, 'xmi.version', '1.3')
        content = doc.createElementNS(NO_NS, 'XMI.content')
        root.appendChild(content)    
        n1 = self.createModelElement('Model_Management.Model', project)
        add_std_attrs(n1)
        content.appendChild(n1)
        content = doc.createElementNS(NO_NS, '%sNamespace.ownedElement'%_CORE)
        n1.appendChild(content)
        return root, content

    def open_package(self, node):
        doc = self._doc
        el = self.createModelElement('Model_Management.Package', node)
        add_std_attrs(el, 'false', 'false')
        content = doc.createElementNS(NO_NS, '%sNamespace.ownedElement'%_CORE)
        el.appendChild(content)    
        return el, content
    
    def open_class_def(self, node):
        """ open a xmi class definition """
        doc = self._doc
        interface = (node.type == 'interface')
        if interface:
            el = self.createModelElement('%sInterface'%_CORE, node)
            add_std_attrs(el, node.final, node.abstract)
        elif node.signal == 'true':
            el = self.createModelElement('Behavioral_Elements.Common_Behavior.Signal', node)
            add_std_attrs(el, node.final, node.abstract)
        else:
            el = self.createModelElement('%sClass'%_CORE, node)
            add_std_attrs(el, node.final, node.abstract)
            add_value(el, '%sClass.isActive'%_CORE, 'false')
        spec, klass = None, []
        if node.subclasses:
            if interface:
                pe = add_me_dependency(el, client=0)
                a_name = '%sAbstraction'%_CORE
            else:
                # delay attachment to be conform to the dtd
                spec = pe = self._doc.createElementNS(
                    NO_NS, '%sGeneralizableElement.specialization'%_CORE)
                a_name = '%sGeneralization'%_CORE
            for p in node.subclasses:
                if isinstance(p, Klass):
                    add_idref(pe, a_name, self.get_x_ref(
                        node, interface))            
        if node.parents:
            for p in node.parents:
                if isinstance(p, Klass):
                    if p.type == 'interface':
                        pe = add_me_dependency(el, client=1)
                        add_idref(pe, '%sAbstraction'%_CORE,
                                             self.get_x_ref(p, 1, node))
                    else:
                        klass.append(p)
        if klass:
            pe = doc.createElementNS(
                NO_NS, '%sGeneralizableElement.generalization'%_CORE)
            el.appendChild(pe)
            for p in klass:
                add_idref(pe, '%sGeneralization'%_CORE,
                                     self.get_x_ref(p, 0, node))
        if spec: el.appendChild(spec)
        
        if node.childs:
            fe = self._doc.createElementNS(NO_NS, '%sClassifier.feature'%_CORE)
            el.appendChild(fe)
            funcs = []
            for child in node.childs:
                if self._filter:
                    if self._filter & _PROTECTED and child.visibility == 'protected':
                        continue
                    if self._filter & _PRIVATE and child.visibility == 'private':
                        continue
                    if child.name == '__init__':
                        if self._filter & _CONSTRUCTOR:
                            continue
                    elif self._filter & _SPECIAL and child.special:
                        continue
                if child.TYPE == 'attribute':
                        fe.appendChild(self.open_attr(child))
                else:
                    funcs.append(child)
            if funcs:
                for child in funcs:
                    fe.appendChild(self.open_func_def(child))
        return el
    
    def open_attr(self, node):
        """ open a xmi class attribute definition """
        attr = self.createModelElement('%sAttribute'%_CORE, node)
        if hasattr(node, 'type') and node.type == 'class':
            add_value(attr, '%sFeature.ownerScope'%_CORE, 'classifier')
        if hasattr(node, 'default') and not isinstance(node.default, AbstractBase) and \
           len(node.default) < MAX_CHAR:
            doc = self._doc
            el = doc.createElementNS(NO_NS, '%sAttribute.initialValue'%_CORE)
            attr.appendChild(el)
            exp = 'Foundation.Data_Types.Expression'
            el2 = id_element(doc, exp, self.generateId(), el)
            el = doc.createElementNS(NO_NS, '%s.language'%exp)
            el2.appendChild(el)
            el.appendChild(doc.createTextNode('Java'))
            el = doc.createElementNS(NO_NS, '%s.body'%exp)
            el2.appendChild(el)
            el.appendChild(doc.createTextNode(node.default))
        return attr
    
    def open_func_def(self, node):
        """ open a xmi class function definition """
        op = self.createModelElement('%sOperation'%_CORE, node)
        if hasattr(node, 'type') and node.type == 'class':
            add_value(op, '%sFeature.ownerScope'%_CORE, 'classifier')
        add_value(op, '%sBehavioralFeature.isQuery'%_CORE, 'false')
        add_std_attrs(op, node.final)
        if node.childs:
            fe = self._doc.createElementNS(
                NO_NS, '%sBehavioralFeature.parameter'%_CORE)
            op.appendChild(fe)
            para = '%sParameter'%_CORE
            for child in node.childs:
                p = self.createModelElement(para, child, uuid=None)
                add_value(p, '%s.kind'%para, "in")
                fe.appendChild(p)
        return op
    
    def open_association(self, attr_n):
        """ open a xmi association definition """
        if hasattr(attr_n, '_resolved_'):
            return
        attr_n._resolved_ =1
        doc = self._doc
        id = self.generateId()
        el = id_element(doc, '%sAssociation'%_CORE, id)
        el.setAttributeNS(NO_NS, 'xmi.uuid', self.formatUUID(id))
        add_value(el, '%sModelElement.isSpecification'%_CORE, 'false')
        add_std_attrs(el, 'false', 'false')
        me = doc.createElementNS(NO_NS, '%sAssociation.connection'%_CORE)
        el.appendChild(me)
        for child in attr_n.default.childs:
            if child.TYPE == 'attribute' and hasattr(child, 'default') and \
               child.default == attr_n:
                child._resolved_ = 1
                nav = 'true'
                break
        else: nav = 'false'
        me.appendChild(self.open_association_end(attr_n.name, attr_n.get_class()))
        me.appendChild(self.open_association_end(attr_n.default.name,
                                                 attr_n.default.get_class(),
                                                 navigable=nav))
        return el
    
    def open_association_end(self, name, end, lower=1, upper=1,
                             navigable='true', aggreg='none'):
        """ open a xmi association end definition """
        doc = self._doc
        el = id_element(doc, '%sAssociationEnd'%_CORE, self.generateId())
        me = doc.createElementNS(NO_NS, '%sModelElement.name'%_CORE)
        el.appendChild(me)
        me.appendChild(doc.createTextNode(name))
        ae = '%sAssociationEnd.'%_CORE
        add_value(el, '%sModelElement.isSpecification'%_CORE, 'false')
        add_value(el, '%sisNavigable'%ae, navigable)
        add_value(el, '%saggregation'%ae, aggreg)
        mul = 'Foundation.Data_Types.Multiplicity'
        n = doc.createElementNS(NO_NS, '%smultiplicity'%ae)
        el.appendChild(n)
        mul = 'Foundation.Data_Types.Multiplicity'
        me = id_element(doc, mul, self.generateId(), n)
        range = doc.createElementNS(NO_NS, '%s.range'%mul)
        me.appendChild(range)
        me = id_element(doc, '%sRange'%mul, self.generateId(), range)
        n = doc.createElementNS(NO_NS, '%sRange.lower'%mul)
        n.appendChild(doc.createTextNode(str(lower)))
        me.appendChild(n)
        n = doc.createElementNS(NO_NS, '%sRange.upper'%mul)
        n.appendChild(doc.createTextNode(str(upper)))
        me.appendChild(n)
        me = doc.createElementNS(NO_NS, '%stype'%ae)
        el.appendChild(me)
        if end.type == 'interface':
            add_idref(me, '%sInterface'%_CORE, end.id)
        elif end.signal == 'true':
            add_idref(me, 'Behavioral_Elements.Common_Behavior.Signal', end.id)
        else:
            add_idref(me, '%sClass'%_CORE, end.id)
        return el
        
    def get_x_ref(self, root, interface=None, dependency=None):
        """ manage cross links between two objects (generalization, implements)
        root is the ancestor node.
        This function is called two time for each link:
          _ one with dependency=None
          _ one with dependency references the specialization object.
        We don't know the order :(
        """
        doc = self._doc
        if not hasattr(root, '_tmp_xmi_ref'):
            root._xmi_ref = {}
            root._last_xmi_ref, root._tmp_xmi_ref = [], []
        if dependency is not None and root._tmp_xmi_ref:
            root._xmi_ref[dependency] = e = root._tmp_xmi_ref[-1]
            self.createXref(self._xref_tab[e][0], root, interface, dependency)
            del root._tmp_xmi_ref[-1]
            return e
        elif dependency is None and root._last_xmi_ref:
            e = root._xmi_ref[root._last_xmi_ref[-1]]
            self.createXref(self._xref_tab[e][0], root, interface, dependency)
            del root._last_xmi_ref[-1]
            return e
        ## create dependency
        id = self.generateId()
        if dependency:
            root._xmi_ref[dependency] = id
            root._last_xmi_ref.append(dependency)
        else: 
            root._tmp_xmi_ref.append(id)
        if interface:
            xref = id_element(doc, '%sAbstraction'%_CORE, id)
            # extension
            ext_id = self.generateId()
            ext = id_element(doc, 'Foundation.Extension_Mechanisms.Stereotype', ext_id)
            core = '%sModelElement.' % _CORE
            me = doc.createElementNS(NO_NS, '%sname'%core)
            ext.appendChild(me)
            me.appendChild(doc.createTextNode('realize'))
            add_value(ext, '%sisSpecification'%core, 'false')
            add_std_attrs(ext)
            el = doc.createElementNS(
                NO_NS, 'Foundation.Extension_Mechanisms.Stereotype.extendedElement')
            ext.appendChild(el)
            add_idref(el, 'Foundation.Core.Abstraction', id) 
        else:
            xref = id_element(doc, '%sGeneralization'%_CORE, id)
            ext = None
        xref.setAttributeNS(NO_NS, 'xmi.uuid', self.formatUUID(id))
        add_value(xref, '%sModelElement.isSpecification'%_CORE,
                             'false')
        if not ext is None:
            self._xref_tab[ext_id] = (ext, self._stack[-1])
            ext = doc.createElementNS( NO_NS, '%sModelElement.stereotype'%_CORE)
            xref.appendChild(ext)
            add_idref(
                ext, 'Foundation.Extension_Mechanisms.Stereotype', ext_id)
        self.createXref(xref, root, interface, dependency)
        self._xref_tab[id] = (xref, self._stack[-1])
        #self._stack[0].appendChild(xref)
        return id
    
    ## XMI functions node creation ############################################
    def createModelElement(self, n_name, node, uuid=1):
        """ create a XMI ModelElement node """
        doc = self._doc
        id = int(node.id)
        el = id_element(doc, n_name, id)
        if uuid is not None:
            el.setAttributeNS(NO_NS, 'xmi.uuid', self.formatUUID(id))
        core = '%sModelElement.' % _CORE
        me = doc.createElementNS(NO_NS, '%sname'%core)
        el.appendChild(me)
        me.appendChild(doc.createTextNode(node.name))
        if hasattr(node, 'visibility'):
            add_value(el, '%svisibility'%core, node.visibility)
        add_value(el, '%sisSpecification'%core, 'false')
        return el
    
    # FIXME interface !! beurk beurk
    def createXref(self, node, root, interface, dependency):
        """ create a XMI cross link reference """
        doc = self._doc
        if dependency:
            if not interface:
                d = doc.createElementNS(NO_NS, '%sGeneralization.child'%_CORE)
            else:
                d = doc.createElementNS(NO_NS, '%sDependency.client'%_CORE)
            a_name = '%sClass'%_CORE
            add_idref(d, a_name, dependency.id)
            if node.nodeName == '%sAbstraction'%_CORE:
                node.insertBefore(d, node.firstChild.nextSibling.nextSibling)
            else: 
                node.insertBefore(d, node.firstChild.nextSibling)
        else:
            if not interface:
                d = doc.createElementNS(NO_NS, '%sGeneralization.parent'%_CORE)
                a_name = '%sClass'%_CORE
            else:
                d = doc.createElementNS(NO_NS, '%sDependency.supplier'%_CORE)
                a_name = '%sInterface'%_CORE
            add_idref(d, a_name, root.id) 
            node.appendChild(d)

    def formatUUID(self, id):
        """ format xmi uuid """
        return '%s-%x' % (self._uuid, 0x8000 - id)
    

def format_id(id):
    """ format xmi identifiers """
    return 'xmi-%d' % id

def id_element(doc, name, id, parent=None):
    """ add an XMI id element """
    e = doc.createElementNS(NO_NS, name)
    e.setAttributeNS(NO_NS, 'xmi.id', format_id(id))
    if parent is not None:
        parent.appendChild(e)
    return e

def add_idref(parent, name, id):
    """ add an XMI idref element """
    me_spec = parent.ownerDocument.createElementNS(NO_NS, name)
    me_spec.setAttributeNS(NO_NS, 'xmi.idref', format_id(id))
    parent.appendChild(me_spec)
    
def add_value(parent, n_name, value):
    """ add an XMI value element """
    me_spec = parent.ownerDocument.createElementNS(NO_NS, n_name)
    me_spec.setAttributeNS(NO_NS, 'xmi.value', value)
    parent.appendChild(me_spec)
    
def add_me_dependency(parent, client, before=None):
    if client: pe = parent.ownerDocument.createElementNS(
        NO_NS, '%sModelElement.clientDependency'%_CORE)
    else: pe = parent.ownerDocument.createElementNS(
        NO_NS, '%sModelElement.supplierDependency'%_CORE)
    if before is None:
        parent.appendChild(pe)
    else:
        parent.insertBefore(pe, before)
    return pe

def add_dependency(parent, client):
    if client: pe = parent.ownerDocument.createElementNS(
        NO_NS, '%sDependency.client'%_CORE)
    else: pe = parent.ownerDocument.createElementNS(
        NO_NS, '%sDependency.supplier'%_CORE)
    parent.appendChild(pe)
    return pe

def add_std_attrs(ge, final='false', abstract='false'):
    """ add some standards attributes to a node """
    if ge.nodeName == '%sOperation'%_CORE:
        elmt = '%sOperation' % _CORE
    else:
        elmt = '%sGeneralizableElement' % _CORE
    add_value(ge, '%s.isRoot'%(elmt), 'false')
    add_value(ge, '%s.isLeaf'%(elmt), final)
    add_value(ge, '%s.isAbstract'%(elmt), abstract)
    
USAGE = """
xmi_uml parse python sources files and extract an XMI UML file from them

Usage: xmi_uml [options] <file>...

Options: _ -h/--help, display this help message and exit
         _ -p/--project <name>, set project name to <name>
         _ -o/--output <file>, write XMI to <file>
"""
import sys
def Run(args):
    import getopt
    (opt, args) = getopt.getopt(args, 'hp:o:', ['help', 'project=', 'output='])
    project, dest = None, sys.stdout
    for name, val in opt:
        if name == '-h' or name == '--help':
            print >>sys.stderr, USAGE
            return
        elif name == '-p' or name == '--project':
            project = val
        elif name == '-o' or name == '--output':
            dest = open(val, 'w+')
    if not project:
        print >>sys.stderr, 'You must specify a project name with the -p/--project option'
        return
    from pyparser import PyParser
    parser = PyParser()
    p = parser.parse_all(args, project)
    p.resolve_hierarchy()
    print p
    k = Klass2DomXmi()
    d = k.parse(p, parser.id_count)
    from xml.dom.ext import PrettyPrint
    PrettyPrint(d.ownerDocument, dest)
    
if __name__ == "__main__":
    Run(sys.argv[1:])



